Actor subclass: #ReplaceActor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Replace-Ext'!
ReplaceActor comment:
'Class ReplaceActor provides behavior replacement messages.
This allows behavior replacement messages to be computed externally and sent to the actor.
This is useful when the replacement behavior needs some non local information to be computed and therefore has to be computed externally by another actor.
Meanwhile the actor becomes insensitive (because it cannot accept any message with no behavior) while waiting (passively) for its replacement behavior.

Class ReplaceActor is a subclass of kernel class Actor.
Method replace: directly handles behavior replacement messages at the actor level (and not at the behavior level). This is necessary because while insensitive the actor may not accept any message!!

Example of use is the management of overdraft. See class ProtectedBankAccount in category Actalk-Ext-Replace-Ext-Ex.
Also immediate extensions are implementations of futures, see class FutureActor.'!


!ReplaceActor methodsFor: 'behavior replacement'!

replace: replacementBehavior
	"Compute behavior replacement messages.
	On one hand effect is the same as for a standard behavior replacement method requested by a behavior,
	on the other hand a replacement behavior message is processed by the actor
	and consequently may be sent externally by another actor."

	self setBehavior: replacementBehavior! !

ReplaceActor subclass: #FutureActor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Replace-Ext'!
FutureActor comment:
'Class FutureActor implements futures (that is objects/actors whose behavior/value has not been assigned yet) simply as a newly created actor with no behavior yet.
In other words a FutureActor is simply a born insensitive actor.
It will buffer all messages until its real nature (behavior) gets computed (externally and assigned through a replacement behavior message).

Actually there is no need for any implementation. No instance variable nor method is defined!!
Its superclass ReplaceActor could be used as well. But we define class FutureActor as a way to clearly identify the use of futures.

Note that the limitation of this model of future is that it value when computed may only be defined through a behavior.
Because Actalk is not fully uniform (not everything is an actor), basic entities such as numbers cannot be easily defined through behaviors. A future actor cannot become such primitive objects.
Future actors are useful to implement external actors, that is, in the terminology of Gul Agha, actors which provide interfacing with other program modules. We may provide the actual connection at execution time, meanwhile future actors will buffer all messages.

One alternative to this trivial implementation relies on more traditional objects and assignment, as well as semaphores. This could be found in category Actalk-Ext-Abcl and will be used to implement eager type (also called future type) message passing.'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

FutureActor class
	instanceVariableNames: ''!


!FutureActor class methodsFor: 'example'!

example
	"FutureActor example"

	FutureActor new
		deposit: 300 replyTo: Print;
		replace: (BankAccount new balance: 200);
		withdraw: 400 replyTo: Print;
		balanceAndReplyTo: Print! !

ReplaceActorBehavior subclass: #AcceptReplaceActorBehavior
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Replace-Ext'!
AcceptReplaceActorBehavior comment:
'Class AcceptReplaceActorBehavior is a subclass of class ReplaceActorBehavior and provides control over acceptance of messages.
Basically it looks only for messages that it knows (selectors of defined methods). This is useful to disable method acceptance (without relying on trapping errors and resending messages as in class NullBoundedBuffer).
Method nextMessage is redefined to look for messages that the class knows (defines or inherits).
See class AcceptNullBoundedBuffer in category Actalk-Ext-Replace-Ext-Ex as the resulting improved version of the bounded buffer example.'!


!AcceptReplaceActorBehavior methodsFor: 'message handling'!

nextMessage
	"Look into the mailbox for the first message matching one of defined method selectors.
	Ignore messages not matching any of them.
	Blocking and suspend until finding any matching one.
	(This means it suspends onto incoming messages as long as necessary)."

	"Finding all defined method selectors (within script protocol) is not optimized
	as collecting them is redone for each message acceptance although it is constant!!
	This could be cached and set somewhere within an instance variable of the class or metaclass."

	^aself mailBox firstMessageAmong: self class allScriptSelectors! !
