CountersActorBehavior subclass: #ASCActorBehavior
	instanceVariableNames: 'enabledSelectors '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-ASC'!
ASCActorBehavior comment:
'Class ASCActorBehavior implements a mixed synchronization model.
ASC stands for Abstract State and Synchronization Counters.
This means using both state transitions to specify state synchronization and activation conditions (guards based on synchronization counters) to specify activation synchronization.
As a result specification is more modular and reusable. However these two synchronization dimensions are not fully orthogonal. (See example of classes ASCGetRearBoundedBuffer and ASCGetRear2BoundedBuffer).
This model has been proposed by Laurent Thomas at PARLE92 Conference.
Class ASCActorBehavior is a combination of synchronization classes StateActorBehavior and CountersActorBehavior. However because there is no multiple inheritance, we choose the most complex class, CountersActorBehavior as the superclass, and we recopy the state transition part part. Meanwhile we need to be careful about the combination, because we have to ensure atomicity of checking of conditions and method completion. (This is actually easily specified through simple refinement of method checking the guard to also include the enabling check.)'!


!ASCActorBehavior methodsFor: 'initialize'!

actalkPrivateInitialize

	super actalkPrivateInitialize.	"Create the three counters dictionaries."
	self enabledSelectors: (self perform: self initialAbstractState)	"Compute and assign the initial enabled selectors."! !

!ASCActorBehavior methodsFor: 'accessing'!

enabledSelectors: arrayOfSelectors

	enabledSelectors := arrayOfSelectors! !

!ASCActorBehavior methodsFor: 'guard testing'!

satisfyGuardSelector: selector arguments: argumentsArray
	"Check if the message belongs to current set of enabled selectors
	AND if the guard evaluates to true.
	Both checks must be done within the same atomic block."

	^(enabledSelectors includes: selector)
		and: [super satisfyGuardSelector: selector arguments: argumentsArray]! !

!ASCActorBehavior methodsFor: 'synchro events'!

synchroComplete: aMessage

	super synchroComplete: aMessage.
	self enabledSelectors: (self computeNextEnabledSelectorsAfter: aMessage selector)! !

!ASCActorBehavior methodsFor: 'abstract states'!

initialAbstractState
	"Return the initial abstract state.
	Mandatory. Must be supplied by the class or one of its superclass.
	Must return some abstract state symbol/selector."

	^self subclassResponsability! !

!ASCActorBehavior methodsFor: 'state transition'!

computeNextEnabledSelectorsAfter: selector
	"Compute (the array of) next enabled selectors.
	First compute next abstract state (by calling method nextAbstractStateAfter:)
	and then perform it to obtain array of next enabled selectors.
	Note that the knowledge of the selector may be useful so it is included as an argument."

	^self perform: (self nextAbstractStateAfter: selector)!

nextAbstractStateAfter: selector
	"Specify and compute next abstract state.
	Mandatory. Must be supplied by the class or one of its superclass.
	Must return some abstract state symbol/selector."

	^self subclassResponsability! !
