GuardsActorBehavior subclass: #GuardsPhilosopher
	instanceVariableNames: 'leftForkPosition rightForkPosition table numberSteps '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Guards-Ex'!
GuardsPhilosopher comment:
'Class GuardsPhilosopher implements the dining philosophers example with guards.
As opposed to the solution with POOL explicit message acceptance (see class PoolPhilosopher) which needed to restrict the number of philosophers accessing the table in order to esure deadlock avoidance, this description does not need to.
The solution makes use of a class GuardsTable to centralize access to the forks in order to ensure that a philosopher picks up his two forks at once (atomically) if they are both available.
However this slightly restricts the ideal concurrency of accesses to forks, that is it serializes accesses to forks even if two philosophers use different forks.'!


!GuardsPhilosopher methodsFor: 'initialize'!

position: anInteger n: n table: aTable numberSteps: steps

	leftForkPosition := anInteger.
	rightForkPosition := (leftForkPosition \\ n) + 1.
	table := aTable.
	numberSteps := steps! !

!GuardsPhilosopher methodsFor: 'script'!

start
	"Activity of a philosopher."

	numberSteps timesRepeat:
		[self think.
		Transcript show: self printString , ' wants to eat and requests forks '
			, leftForkPosition printString , ' and ' , rightForkPosition printString; cr.
		"He requests his forks. Note that he waits (value) for the acknowledgement."
		(table pickUp: leftForkPosition and: rightForkPosition) value.
		self eat.
		"Note that when releasing forks, he does not need to wait for acknowledgement to proceed."
		table putDown: leftForkPosition and: rightForkPosition.
		Transcript show: self printString , ' has finished eating and releases forks '
			, leftForkPosition printString , ' and ' , rightForkPosition printString; cr]! !

!GuardsPhilosopher methodsFor: 'guards'!

guardOFstart

	^true! !

!GuardsPhilosopher methodsFor: 'private routines'!

eat

	| delay |
	delay := (Random new next * 5) rounded + 1.
	Transcript show: self printString , ' eats during ' , delay printString , ' seconds.'; cr.
	(Delay forSeconds: delay) wait!

think

	| delay |
	delay := (Random new next * 5) rounded + 1.
	Transcript show: self printString , ' thinks during ' , delay printString , ' seconds.'; cr.
	(Delay forSeconds: delay) wait! !

!GuardsPhilosopher methodsFor: 'printing'!

printOn: aStream

	aStream nextPutAll: self class name , leftForkPosition printString! !

CheckGuardsActorBehavior subclass: #GuardsCheck
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Guards-Ex'!


!GuardsCheck methodsFor: 'script'!

bar: b

	^b!

foo

	^0!

foo: a bar: b

	^a+b! !

!GuardsCheck methodsFor: 'guards'!

guardOFbar: b

	^b isNil! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

GuardsCheck class
	instanceVariableNames: ''!


!GuardsCheck class methodsFor: 'counter example'!

example
	"GuardsCheck example"

	"Should produce warnings and default guards when absent."

	self initialize! !

GuardsActorBehavior subclass: #GuardsBoundedBuffer
	instanceVariableNames: 'contents maxSize '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Guards-Ex'!


!GuardsBoundedBuffer methodsFor: 'initialize'!

initialize: anInteger

	maxSize := anInteger.
	contents := BoundedBufferObject new maxSize: maxSize! !

!GuardsBoundedBuffer methodsFor: 'script'!

get

	^contents removeFirst!

put: item

	contents addLast: item! !

!GuardsBoundedBuffer methodsFor: 'guards'!

guardOFget
	"Not empty."

	^contents isEmpty not!

guardOFput: item
	"Not full."

	^contents isFull not! !

!GuardsBoundedBuffer methodsFor: 'printing'!

printOn: aStream
	"Print as its class name, then its contents."

	super printOn: aStream.
	contents printOn: aStream! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

GuardsBoundedBuffer class
	instanceVariableNames: ''!


!GuardsBoundedBuffer class methodsFor: 'instance creation'!

new: size

	^self new initialize: size! !

!GuardsBoundedBuffer class methodsFor: 'example'!

exampleSize: maxSize numberItems: numberItems speedRatio: ratio
	"GuardsBoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 4"
	"GuardsBoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 1/4"

	| buffer producer consumer |
	buffer := (GuardsBoundedBuffer new: maxSize) actor.
	producer := (Producer new buffer: buffer delay: 1) actor.
	consumer := (Consumer new buffer: buffer delay: ratio) actor.
	producer runPut: numberItems.
	consumer runGet: numberItems! !

GuardsBoundedBuffer subclass: #GuardsGet2BoundedBuffer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Guards-Ex'!


!GuardsGet2BoundedBuffer methodsFor: 'script'!

get2

	^Array with: contents removeFirst with: contents removeFirst! !

!GuardsGet2BoundedBuffer methodsFor: 'guards'!

guardOFget2
	"At least one element."

	^contents size > 1! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

GuardsGet2BoundedBuffer class
	instanceVariableNames: ''!


!GuardsGet2BoundedBuffer class methodsFor: 'example'!

exampleSize: maxSize numberItems: numberItems speedRatio: ratio
	"GuardsGet2BoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 2"

	| buffer producer consumer |
	buffer := (GuardsGet2BoundedBuffer new: maxSize) actor.
	producer := (Producer new buffer: buffer delay: ratio) actor.
	consumer := (Consumer new buffer: buffer delay: 1) actor.
	producer runPut: numberItems.
	consumer runGet2: numberItems! !

GuardsBoundedBuffer subclass: #GuardsGgetBoundedBuffer
	instanceVariableNames: 'afterPut '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Guards-Ex'!


!GuardsGgetBoundedBuffer methodsFor: 'script'!

get

	afterPut := false.
	^super get!

gget

	^self get!

put: item

	afterPut := true.
	contents addLast: item.! !

!GuardsGgetBoundedBuffer methodsFor: 'guards'!

guardOFgget
	"Not empty and not after a put."

	^(self guardOFget) & afterPut not! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

GuardsGgetBoundedBuffer class
	instanceVariableNames: ''!


!GuardsGgetBoundedBuffer class methodsFor: 'example'!

exampleSize: maxSize numberItems: numberItems speedRatio: ratio
	"GuardsGgetBoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 1"

	| buffer producer consumer |
	buffer := (GuardsGgetBoundedBuffer new: maxSize) actor.
	producer := (Producer new buffer: buffer delay: 1) actor.
	consumer := (Consumer new buffer: buffer delay: ratio) actor.
	producer runPut: numberItems.
	consumer runGget: numberItems! !

Guards1ActorBehavior subclass: #Guards1BoundedBuffer
	instanceVariableNames: 'contents maxSize '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Guards-Ex'!


!Guards1BoundedBuffer methodsFor: 'initialize'!

initialize: anInteger

	maxSize := anInteger.
	contents := BoundedBufferObject new maxSize: maxSize! !

!Guards1BoundedBuffer methodsFor: 'script'!

get

	^contents removeFirst!

put: item

	contents addLast: item! !

!Guards1BoundedBuffer methodsFor: 'guards'!

guardOFget
	"Not empty."

	^contents isEmpty not!

guardOFput: item
	"Not full."

	^contents isFull not! !

!Guards1BoundedBuffer methodsFor: 'printing'!

printOn: aStream
	"Print as its class name, then its contents."

	super printOn: aStream.
	contents printOn: aStream! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Guards1BoundedBuffer class
	instanceVariableNames: ''!


!Guards1BoundedBuffer class methodsFor: 'instance creation'!

new: size

	^self new initialize: size! !

!Guards1BoundedBuffer class methodsFor: 'example'!

exampleSize: maxSize numberItems: numberItems speedRatio: ratio
	"Guards1BoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 4"
	"Guards1BoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 1/4"

	| buffer producer consumer |
	buffer := (Guards1BoundedBuffer new: maxSize) actor.
	producer := (Producer new buffer: buffer delay: 1) actor.
	consumer := (Consumer new buffer: buffer delay: ratio) actor.
	producer runPut: numberItems.
	consumer runGet: numberItems! !

GuardsActorBehavior subclass: #GuardsTable
	instanceVariableNames: 'n numberSteps philosophers forksAvailability '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Guards-Ex'!
GuardsTable comment:
'Class GuardsTable implements the shared ressource of philosophers, that is forks placed onto the table. See class GuardsPhilosopher comment.'!


!GuardsTable methodsFor: 'initialize'!

initialize: anInteger numberSteps: steps

	n := anInteger.
	numberSteps := steps!

makeForksAvailability
	"Create and initialize the array containing the availability of each fork."

	forksAvailability := Array new: n withAll: true!

makePhilosophers
	"Create and initialize the n philosophers."

	philosophers := Array new: n.
	1 to: n do:
		[:i | philosophers at: i put:
			(GuardsPhilosopher new
				position: i
				n: n
				table: aself
				numberSteps: numberSteps)
					actor]! !

!GuardsTable methodsFor: 'script'!

pickUp: i and: j

	forksAvailability at: i put: false;
				at: j put: false!

putDown: i and: j

	forksAvailability at: i put: true;
				at: j put: true!

start

	self makeForksAvailability.
	self makePhilosophers.
	philosophers do:
		[:philosopher | philosopher start]! !

!GuardsTable methodsFor: 'guards'!

guardOFpickUp: i and: j
	"Both forks must be available."

	^(forksAvailability at: i) & (forksAvailability at: j)!

guardOFputDown: i and: j
	"Always free to release forks."

	^true!

guardOFstart

	^true! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

GuardsTable class
	instanceVariableNames: ''!


!GuardsTable class methodsFor: 'example'!

example: n
	"GuardsTable example: 5"

	(GuardsTable new initialize: n numberSteps: 3) actor start! !
